/*
 * Siemens USB-MPI Serial USB driver
 *
 * Copyright (C) 2005 Thomas Hergenhahn <thomas.hergenhahn@suse.de>
 * Copyright (C) 2005,2008 Greg Kroah-Hartman <gregkh@suse.de>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License version
 *	2 as published by the Free Software Foundation.
 */

//#define DEBUG

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/version.h>
#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/slab.h>
#include <linux/tty_flip.h>
#include <linux/module.h>
#include <linux/usb.h>
#include <linux/usb/serial.h>

#define DRIVER_AUTHOR "Marco Gaibazzi"
#define DRIVER_DESC "Driver for Custom Printers USB virtual com"

#define DRIVER_VERSION		"1.00"
#define DRIVER_SCODE		"01DV0000000162"
#define DRIVER_CUSTOM_NAME	"Custom Virtual Com"

#define BIT0	(1<<0)
#define BIT1	(1<<1)

#define SET_LINE_CODING         0x20
#define GET_LINE_CODING         0x21
#define SET_CONTROL_LINE_STATE  0x22

static __u16 vendor = 0;
static __u16 product = 0;

module_param(vendor, ushort, 0);
MODULE_PARM_DESC(vendor, "User specified USB idVendor");

module_param(product, ushort, 0);
MODULE_PARM_DESC(product, "User specified USB idProduct");

#define USB_CUSTOM_VCOM_INTCLASS	0xff
#define USB_CUSTOM_VCOM_INTSUBCLASS	0x01
#define USB_CUSTOM_VCOM_INTPROTOCOL	0x00




struct custom_private {
        spinlock_t lock;        /* protects the following flags */
        bool rts; //tx
        bool cts; //rx
        bool dtr; //tx
        bool dsr; //rx
};



static struct usb_device_id id_table[] = {
	/* KPM150H */
	{ USB_DEVICE_AND_INTERFACE_INFO(0x0dd4, 0x8203, USB_CUSTOM_VCOM_INTCLASS, USB_CUSTOM_VCOM_INTSUBCLASS, USB_CUSTOM_VCOM_INTPROTOCOL) },
	/* KPM180H */
	{ USB_DEVICE_AND_INTERFACE_INFO(0x0dd4, 0x8208, USB_CUSTOM_VCOM_INTCLASS, USB_CUSTOM_VCOM_INTSUBCLASS, USB_CUSTOM_VCOM_INTPROTOCOL) },
	/* KPM300H */
	{ USB_DEVICE_AND_INTERFACE_INFO(0x0dd4, 0x8175, USB_CUSTOM_VCOM_INTCLASS, USB_CUSTOM_VCOM_INTSUBCLASS, USB_CUSTOM_VCOM_INTPROTOCOL) },
	/* Q3 */
	{ USB_DEVICE_AND_INTERFACE_INFO(0x0dd4, 0x818e, USB_CUSTOM_VCOM_INTCLASS, USB_CUSTOM_VCOM_INTSUBCLASS, USB_CUSTOM_VCOM_INTPROTOCOL) },
	/* VKP80III */
	{ USB_DEVICE_AND_INTERFACE_INFO(0x0dd4, 0x8205, USB_CUSTOM_VCOM_INTCLASS, USB_CUSTOM_VCOM_INTSUBCLASS, USB_CUSTOM_VCOM_INTPROTOCOL) },
	/* K3 */
	{ USB_DEVICE_AND_INTERFACE_INFO(0x0dd4, 0x820a, USB_CUSTOM_VCOM_INTCLASS, USB_CUSTOM_VCOM_INTSUBCLASS, USB_CUSTOM_VCOM_INTPROTOCOL) },
	{ },
	{ },
};
MODULE_DEVICE_TABLE(usb, id_table);


static int send_control_msg(struct usb_serial_port *port, u8 request, u16 value)
{
        struct usb_serial *serial = port->serial;
        int retval;

	dev_dbg(&port->dev, "send_control_msg request=0x%x value=0x%x\n", request, value);

        /* Send the message to the vendor control endpoint
         * of the connected device */
        retval = usb_control_msg(serial->dev, usb_sndctrlpipe(serial->dev, 0),
                                request,
                                USB_DIR_OUT|USB_TYPE_CLASS,
                                value, 0, NULL, 0, 1000);

        return retval;
}

static int send_pinout_msg(struct usb_serial_port *port, bool rts, bool dtr)
{
	u16 value;
	int ret;

	value = 0;
	if ( rts )
		value |= BIT1;
	if ( dtr )
		value |= BIT0;
        ret = send_control_msg(port, SET_CONTROL_LINE_STATE, value);
	return ret;
}


static void custom_process_data_packet(struct usb_serial_port *port,
                                        const unsigned char *buf, size_t len)
{
	char *ch = (char *)buf;
	int i;
        struct tty_struct *tty_s_ptr;
#if LINUX_VERSION_CODE > KERNEL_VERSION(3, 8, 0)
        struct tty_port *tty;
#else
        struct tty_struct *tty;
#endif

	if ( !len )
		return;

#if LINUX_VERSION_CODE > KERNEL_VERSION(3, 8, 0)
	tty = &port->port;
	tty_s_ptr = tty_port_tty_get(&port->port);
#else
        tty = tty_port_tty_get(&port->port);
	tty_s_ptr = tty;
#endif
        if ( (!tty) || (!tty_s_ptr) )
                return;

        /* The per character mucking around with sysrq path it too slow for
           stuff like 3G modems, so shortcircuit it in the 99.9999999% of cases
           where the USB serial is not a console anyway */
        if (!port->port.console || !port->sysrq)
                tty_insert_flip_string(tty, ch, len);
        else {
                for (i = 0; i < len; i++, ch++) {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 35)
                        if (!usb_serial_handle_sysrq_char(port, *ch))
#else
                        if (!usb_serial_handle_sysrq_char(tty, port, *ch))
#endif
                                tty_insert_flip_char(tty, *ch, TTY_NORMAL);
                }
        }

        //tty_insert_flip_string(tty, buf, len);
        tty_flip_buffer_push(tty);
        tty_kref_put(tty_s_ptr);
}


static void custom_process_status_packet(struct usb_serial_port *port, unsigned char status)
{
	struct custom_private *priv = usb_get_serial_port_data(port);
        unsigned long flags;

	dev_dbg(&port->dev, "status=0x%x\n", status);

        spin_lock_irqsave(&priv->lock, flags);
	priv->cts = (status & BIT0) ?  true : false;
	priv->dsr = (status & BIT1) ? true : false;
        spin_unlock_irqrestore(&priv->lock, flags);

}

static void custom_process_read_urb(struct urb *urb)
{
        struct usb_serial_port *port = urb->context;
        const unsigned char *hdr = urb->transfer_buffer;
        const unsigned char *data = hdr + 1;
        size_t data_len = urb->actual_length - 1;

	dev_dbg(&port->dev, "custom_process_read_urb urb->actual_length=%d\n", urb->actual_length);
        if (urb->actual_length < 1) {
                dev_dbg(&port->dev, "malformed packet received: %d bytes\n",
                                                        urb->actual_length);
                return;
        }
        /*
         * Data from the device comes with a 1 byte header:
         *
         * <0x00>data...
         */
        custom_process_status_packet(port, *hdr);
        custom_process_data_packet(port, data, data_len);
}

static void custom_init_signals(struct usb_serial_port *port)
{
        struct custom_private *priv = usb_get_serial_port_data(port);
        unsigned long flags;
	bool rts, dtr;

        spin_lock_irqsave(&priv->lock, flags);
        priv->rts = false;
        priv->cts = false;
        priv->dsr = false;
        priv->dtr = true;
	rts = priv->rts;
	dtr = priv->dtr;
        spin_unlock_irqrestore(&priv->lock, flags);

	send_pinout_msg(port, rts, dtr);

        /* clear the halt status of the enpoint */
        usb_clear_halt(port->serial->dev, port->read_urb->pipe);
}


static int custom_port_probe(struct usb_serial_port *port)
{
        struct custom_private *priv;

	dev_dbg(&port->dev, "custom_port_probe\n");

        priv = kzalloc(sizeof(*priv), GFP_KERNEL);
        if (!priv)
                return -ENOMEM;

        spin_lock_init(&priv->lock);

        usb_set_serial_port_data(port, priv);

	custom_init_signals(port);

        return 0;
}

static int custom_port_remove(struct usb_serial_port *port)
{
        struct custom_private *priv = usb_get_serial_port_data(port);

        kfree(priv);

        return 0;
}

static int custom_open(struct tty_struct *tty, struct usb_serial_port *port)
{
        int res;

	//custom_init_signals(port);

        res = usb_serial_generic_open(tty, port);
        if (!res)
                return res;

        return res;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 38)
static int custom_tiocmget(struct tty_struct *tty)
#else
static int custom_tiocmget(struct tty_struct *tty, struct file *file)
#endif
{
        struct usb_serial_port *port = tty->driver_data;
        struct custom_private *priv = usb_get_serial_port_data(port);
        unsigned long flags;
        int result = 0;

        spin_lock_irqsave(&priv->lock, flags);
        if (priv->rts)
                result |= TIOCM_RTS;
        if (priv->cts)
                result |= TIOCM_CTS;
        if (priv->dtr)
                result |= TIOCM_DTR;
        if (priv->dsr)
                result |= TIOCM_DSR;
        spin_unlock_irqrestore(&priv->lock, flags);

	dev_dbg(&port->dev, "custom_tiocmget result = 0x%x\n", result);

        //dev_dbg(&port->dev, "%s - %x\n", __func__, result);
        return result;
}

#if 0
static void custom_dtr_rts(struct usb_serial_port *port, int on)
{
        struct custom_private *priv = usb_get_serial_port_data(port);
        unsigned long flags;
        bool rts;
        bool dtr;
        bool changedRTS = false;
        bool changedDTR = false;
        int ret;


        spin_lock_irqsave(&priv->lock, flags);

        rts = priv->rts;
        if (on)
                priv->rts = true;
	else
                priv->rts = false;
        changedRTS = rts ^ priv->rts;
        rts = priv->rts;

        dtr = priv->dtr;
        if (on)
                priv->dtr = true;
	else
                priv->dtr = false;
        changedDTR = dtr ^ priv->dtr;
        dtr = priv->dtr;

        spin_unlock_irqrestore(&priv->lock, flags);

        if ( (!changedRTS) && (!changedDTR) )
                return ;

        /* Send the new RTS/DTR state to the connected device */
	ret = send_pinout_msg(port, rts, dtr);

}
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 38)
static int custom_tiocmset(struct tty_struct *tty,
                           unsigned int set, unsigned int clear)
#else
static int custom_tiocmset(struct tty_struct *tty, struct file *file,
                           unsigned int set, unsigned int clear)
#endif
{
        struct usb_serial_port *port = tty->driver_data;
        struct usb_serial *serial = port->serial;
        struct custom_private *priv = usb_get_serial_port_data(port);
        unsigned long flags;
        bool rts;
        bool dtr;
        bool changedRTS = false;
        bool changedDTR = false;
        int ret;

	dev_dbg(&port->dev, "custom_tiocmset \n");
        /* We only support RTS so we only handle that */
        spin_lock_irqsave(&priv->lock, flags);

        rts = priv->rts;
        if (set & TIOCM_RTS)
                priv->rts = true;
        if (clear & TIOCM_RTS)
                priv->rts = false;
        changedRTS = rts ^ priv->rts;
        rts = priv->rts;

        dtr = priv->dtr;
        if (set & TIOCM_DTR)
                priv->dtr = true;
        if (clear & TIOCM_DTR)
                priv->dtr = false;
        changedDTR = dtr ^ priv->dtr;
        dtr = priv->dtr;

        spin_unlock_irqrestore(&priv->lock, flags);



        if ( (!changedRTS) && (!changedDTR) )
                return 0;

        /* Send the new RTS/DTR state to the connected device */
        mutex_lock(&serial->disc_mutex);
        if (!serial->disconnected)
	{
		ret = send_pinout_msg(port, rts, dtr);
	}
        else
                ret = -ENODEV;
        mutex_unlock(&serial->disc_mutex);

        return ret;
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 4, 0)
static struct usb_driver custom_usb_vcom_driver = {
        .name =         "customvcom",
        .probe =        usb_serial_probe,
        .disconnect =   usb_serial_disconnect,
        .id_table =     id_table,
        .no_dynamic_id =        1,
};
#endif



static struct usb_serial_driver custom_usb_vcom_device = {
	.driver = {
		.owner =	THIS_MODULE,
		.name =		"customvcom",
	},
	.id_table =		id_table,
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 4, 0)
	.usb_driver =           &custom_usb_vcom_driver,
#endif
	.num_ports =		1,
	.process_read_urb =     custom_process_read_urb,
        .port_probe =           custom_port_probe,
        .port_remove =          custom_port_remove,
	.open =                 custom_open,
        .tiocmget =             custom_tiocmget,
        .tiocmset =             custom_tiocmset,
	//.dtr_rts  = 		custom_dtr_rts,


};

static struct usb_serial_driver * const serial_drivers[] = {
	&custom_usb_vcom_device, NULL
};


static int __init custom_init(void)
{
	int retval;

	printk(KERN_INFO "%s, version: %s\n", DRIVER_CUSTOM_NAME, DRIVER_VERSION);
	printk(KERN_INFO "%s, scode: %s\n", DRIVER_CUSTOM_NAME, DRIVER_SCODE);

        if (vendor > 0 && product > 0) {
                /* Add user specified VID/PID to reserved element of table. */
                int i;
                for (i = 0; id_table[i].idVendor; i++)
                        ;
                id_table[i].match_flags = USB_DEVICE_ID_MATCH_INT_INFO|USB_DEVICE_ID_MATCH_DEVICE;
                id_table[i].idVendor = vendor;
                id_table[i].idProduct = product;
		id_table[i].bInterfaceClass = USB_CUSTOM_VCOM_INTCLASS;
		id_table[i].bInterfaceSubClass = USB_CUSTOM_VCOM_INTSUBCLASS;
		id_table[i].bInterfaceProtocol = USB_CUSTOM_VCOM_INTPROTOCOL;
        }

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 4, 0)
        retval = usb_serial_register(&custom_usb_vcom_device);
        if (retval)
                goto failed_usb_serial_register;
        retval = usb_register(&custom_usb_vcom_driver);
        if (retval)
                goto failed_usb_register;

        return 0;
failed_usb_register:
        usb_serial_deregister(&custom_usb_vcom_device);
failed_usb_serial_register:
        return retval;

#else
        retval = usb_serial_register_drivers(serial_drivers, KBUILD_MODNAME, id_table);
	return retval;
#endif
}

static void __exit custom_exit(void)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 4, 0)
        usb_deregister(&custom_usb_vcom_driver);
        usb_serial_deregister(&custom_usb_vcom_device);
#else
        usb_serial_deregister_drivers(serial_drivers);
#endif
}


module_init(custom_init);
module_exit(custom_exit);


MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
